# Copyright 2022 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exceptions used in the Worker.

Suppose the input dataflow of our model server is as follows:

**bytes** ``->`` *deserialize* ``->`` **data** ``->`` *parse* ``->`` **valid data**

If the raw bytes cannot be successfully deserialized, the `DecodingError`
is raised; if the decoded data cannot pass the validation check (usually
implemented by users), the `ValidationError` should be raised.
"""

from mosec.protocol import HTTPStautsCode


class MosecError(Exception):
    """Mosec basic exception."""

    code: HTTPStautsCode = HTTPStautsCode.INTERNAL_ERROR
    msg: str = "mosec error"


class ClientError(MosecError):
    """Client side error.

    This error indicates that the server cannot or will not process the request
    due to something that is perceived to be a client error. It will return the
    details to the client side with
    `HTTP 400 <https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/400>`__.
    """

    code = HTTPStautsCode.BAD_REQUEST
    msg = "bad request"


class ServerError(MosecError):
    """Server side error.

    This error indicates that the server encountered an unexpected condition
    that prevented it from fulfilling the request. It will return the details
    to the client side with
    `HTTP 500 <https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/500>`__.

    Attention: be careful about the returned message since it may contain some
    sensitive information. If you don't want to return the details, just raise
    an exception that is not inherited from `mosec.errors.MosecError`.
    """

    code = HTTPStautsCode.INTERNAL_ERROR
    msg = "internal error"


class EncodingError(ServerError):
    """Serialization error.

    The `EncodingError` should be raised in user-implemented codes when
    the serialization for the response bytes fails. This error will set
    to status code to
    `HTTP 500 <https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/500>`__
    and show the details in the response.
    """

    msg = "encoding error"


class DecodingError(ClientError):
    """De-serialization error.

    The `DecodingError` should be raised in user-implemented codes
    when the de-serialization for the request bytes fails. This error
    will set the status code to
    `HTTP 400 <https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/400>`__
    in the response.
    """

    msg = "decoding error"


class ValidationError(MosecError):
    """Request data validation error.

    The `ValidationError` should be raised in user-implemented codes,
    where the validation for the input data fails. Usually, it should be
    put after the data de-serialization, which converts the raw bytes
    into structured data. This error will set the status code to
    `HTTP 422 <https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/422>`__
    in the response.
    """

    code = HTTPStautsCode.VALIDATION_ERROR
    msg = "request validation error"


class MosecTimeoutError(BaseException):
    """Exception raised when a MOSEC worker operation times out.

    If a bug in the forward code causes the worker to hang indefinitely, a timeout
    can be used to ensure that the worker eventually returns control to the main
    thread program. When a timeout occurs, the `MosecTimeout` exception is raised.
    This exception can be caught and handled appropriately to perform any necessary
    cleanup tasks or return a response indicating that the operation timed out.

    Note that `MosecTimeout` is a subclass of `BaseException`, not `Exception`.
    This is because timeouts should not be caught and handled in the same way as
    other exceptions. Instead, they should be handled in a separate `except` block
    which isn't designed to break the working loop.
    """

    code = HTTPStautsCode.TIMEOUT_ERROR
    msg = "mosec timeout error"



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MOSEC multiprocessing logging configurations."""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime
from typing import Any, MutableMapping

from mosec.args import get_log_level

MOSEC_LOG_NAME = __name__
MOSEC_LOG_PREFIX = "mosec"
USER_LOG_NAME = "moser.user_log"
USER_LOG_PREFIX = "user"


class MosecFormat(logging.Formatter):
    """Basic mosec log formatter.

    This class uses `datetime` instead of `localtime` to get accurate milliseconds.
    """

    # `%z` will be empty string if the object is naive, so we use `Z` to make it
    # compatible with rfc3339
    default_time_format = "%Y-%m-%dT%H:%M:%S.%fZ "

    def formatTime(self, record: logging.LogRecord, datefmt=None) -> str:
        """Convert to datetime with timezone."""
        time = datetime.fromtimestamp(record.created).utcnow()
        return datetime.strftime(time, datefmt if datefmt else self.default_time_format)


class DebugFormat(MosecFormat):
    """Colorful debug formatter."""

    Purple = "\x1b[35m"
    Blue = "\x1b[34m"
    Green = "\x1b[32m"
    Yellow = "\x1b[33m"
    Red = "\x1b[31m"
    Reset = "\x1b[0m"

    default_format = (
        "%(asctime)s %(levelname)s %(prefix)s::%(filename)s:%(lineno)s"
        " [%(process)d]: %(message)s"
    )

    def __init__(
        self, fmt: str | None = None, datefmt: str | None = None, prefix: str = ""
    ) -> None:
        """Init with `%` style format.

        Args:
            fmt (str): logging message format (% style)
            datefmt (str): datatime format
            prefix (str): prefix of target

        """
        # partially align with rust tracing_subscriber
        self.colors = {
            logging.DEBUG: self.Blue,
            logging.INFO: self.Green,
            logging.WARNING: self.Yellow,
            logging.ERROR: self.Red,
            logging.CRITICAL: self.Purple,
        }
        super().__init__(fmt or self.default_format, datefmt, "%")
        self.prefix = prefix

    def format_level(self, name: str, level: int) -> str:
        """Format a level name with the corresponding color."""
        if level not in self.colors:
            return name
        return f"{self.colors[level]}{name}{self.Reset}"

    def formatMessage(self, record: logging.LogRecord) -> str:
        """Format the logging with colorful level names."""
        fmt = self.default_format.replace(
            "%(levelname)s %(prefix)s",
            f"{self.format_level(record.levelname, record.levelno)} {self.prefix}",
        )
        return fmt % record.__dict__


class JSONFormat(MosecFormat):
    """JSON log formatter."""

    def __init__(
        self, fmt: str | None = None, datefmt: str | None = None, prefix: str = ""
    ) -> None:
        """Init with `%` style format.

        Args:
            fmt (str): logging message format (% style)
            datefmt (str): datatime format
            prefix (str): prefix of target

        """
        super().__init__(fmt, datefmt, "%")
        self.prefix = prefix

    def format(self, record: logging.LogRecord) -> str:
        """Format to a JSON string."""
        # yet another mypy type-check issue
        # https://github.com/python/mypy/issues/2900
        res: MutableMapping[str, Any] = {
            "timestamp": self.formatTime(record),
            "level": record.levelname,
            "fields": {
                "message": record.getMessage(),
            },
            "target": f"{self.prefix}::{record.filename}:{record.funcName}",
        }
        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
            res["fields"]["exc_info"] = record.exc_text
        if record.stack_info:
            res["fields"]["stack_info"] = self.formatStack(record.stack_info)
        return json.dumps(res)


def use_log(level: int, formatter: logging.Formatter, logger_name: str):
    """Configure the global log."""
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    if not logger.handlers:
        stream = logging.StreamHandler()
        stream.setFormatter(formatter)
        logger.addHandler(stream)
    return logger


def use_pretty_log(level: int = logging.DEBUG, prefix: str = "", logger_name: str = ""):
    """Enable colorful log."""
    return use_log(level, DebugFormat(prefix=prefix), logger_name)


def use_json_log(level: int = logging.INFO, prefix: str = "", logger_name: str = ""):
    """Enable JSON format log."""
    return use_log(level, JSONFormat(prefix=prefix), logger_name)


def get_logger():
    """Get the logger used by mosec user for multiprocessing."""
    prefix = USER_LOG_PREFIX
    logger_name = USER_LOG_NAME
    log_level = int(os.environ.get(MOSEC_LOG_NAME, "0"))
    if log_level == logging.DEBUG:
        return use_pretty_log(level=log_level, prefix=prefix, logger_name=logger_name)
    return use_json_log(level=log_level, prefix=prefix, logger_name=logger_name)


def get_internal_logger():
    """Get the logger used by mosec internally for multiprocessing."""
    prefix = MOSEC_LOG_PREFIX
    logger_name = MOSEC_LOG_NAME
    log_level = int(os.environ.get(MOSEC_LOG_NAME, "0"))
    if log_level == logging.DEBUG:
        return use_pretty_log(level=log_level, prefix=prefix, logger_name=logger_name)
    return use_json_log(level=log_level, prefix=prefix, logger_name=logger_name)


def set_logger(level=logging.INFO):
    """Set the environment variable so all the sub-processes can inherit it."""
    os.environ[MOSEC_LOG_NAME] = str(level)


# need to configure it here to make sure all the process can get the same one
set_logger(get_log_level())



# Copyright 2022 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Server-Worker communication protocol."""

import logging
import socket
import struct
import warnings
from enum import IntFlag
from io import BytesIO
from typing import Sequence, Tuple

from mosec.log import get_internal_logger

logger = get_internal_logger()

IPC_LARGE_DATA_SIZE = 1024 * 1024  # set as 1 MB


class HTTPStautsCode(IntFlag):
    """HTTP status code flag."""

    OK = 1  # 200
    BAD_REQUEST = 2  # 400
    VALIDATION_ERROR = 4  # 422
    INTERNAL_ERROR = 8  # 500
    TIMEOUT_ERROR = 16  # 408

    # special one, indicate that it's a SSE
    STREAM_EVENT = 32768


class Protocol:
    """IPC protocol.

    This private class implements the client-side protocol through Unix domain socket
    to communicate with the server.
    """

    # byte formats (https://docs.python.org/3/library/struct.html#format-characters)
    FORMAT_FLAG = "!H"
    FORMAT_BATCH = "!H"
    FORMAT_ID = "!I"
    FORMAT_LENGTH = "!I"
    FORMAT_STATE = "!H"

    # lengths
    LENGTH_TASK_FLAG = 2
    LENGTH_TASK_BATCH = 2
    LENGTH_TASK_ID = 4
    LENGTH_TASK_STATE = 2
    LENGTH_TASK_BODY_LEN = 4

    def __init__(
        self,
        name: str,
        addr: str,
        timeout: float = 2.0,
    ):
        """Initialize the protocol client.

        Args:
            name (str): name of its belonging coordinator.
            addr (str): Unix domain socket address in file system's namespace.
            timeout (float, optional): socket timeout. Defaults to 2.0 seconds.

        """
        self.socket = socket.socket(
            socket.AF_UNIX,
            socket.SOCK_STREAM,
        )
        self.socket.settimeout(timeout)
        self.name = name
        self.addr = addr

    def receive(self) -> Tuple[bytes, Sequence[bytes], Sequence[int], Sequence[bytes]]:
        """Receive tasks from the server."""
        flag = self.socket.recv(self.LENGTH_TASK_FLAG)
        batch_size_bytes = self.socket.recv(self.LENGTH_TASK_BATCH)
        batch_size = struct.unpack(self.FORMAT_BATCH, batch_size_bytes)[0]
        ids, states, payloads = [], [], []
        total_bytes = 0

        while batch_size > 0:
            batch_size -= 1
            id_bytes = self.socket.recv(self.LENGTH_TASK_ID)
            state_bytes = self.socket.recv(self.LENGTH_TASK_STATE)
            length_bytes = self.socket.recv(self.LENGTH_TASK_BODY_LEN)
            length = struct.unpack(self.FORMAT_LENGTH, length_bytes)[0]
            payload = _recv_all(self.socket, length)
            ids.append(id_bytes)
            states.append(struct.unpack(self.FORMAT_STATE, state_bytes)[0])
            payloads.append(payload)
            total_bytes += length

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "%s received %d tasks with ids: %s",
                self.name,
                len(ids),
                struct.unpack("!" + "I" * len(ids), b"".join(ids)),
            )

        if total_bytes > IPC_LARGE_DATA_SIZE:
            warnings.warn(
                f"IPC data ({total_bytes} bytes) is large, "
                "which may affect performance",
                RuntimeWarning,
                stacklevel=2,
            )
        return flag, ids, states, payloads

    def send(
        self,
        flag: int,
        ids: Sequence[bytes],
        states: Sequence[int],
        payloads: Sequence[bytes],
    ):
        """Send results to the server."""
        data = BytesIO()
        data.write(struct.pack(self.FORMAT_FLAG, flag))
        if len(ids) != len(payloads):
            raise ValueError("`ids` have different length with `payloads`")
        batch_size = len(ids)
        data.write(struct.pack(self.FORMAT_BATCH, batch_size))
        if batch_size > 0:
            for task_id, state, payload in zip(ids, states, payloads):
                data.write(task_id)
                data.write(struct.pack(self.FORMAT_STATE, state))
                data.write(struct.pack(self.FORMAT_LENGTH, len(payload)))
                data.write(payload)
        self.socket.sendall(data.getbuffer())
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "%s sent %d(%d) tasks with ids: %s",
                self.name,
                len(ids),
                flag,
                struct.unpack("!" + "I" * len(ids), b"".join(ids)),
            )

    def open(self):
        """Open the socket connection."""
        self.socket.connect(self.addr)
        logger.info("%s socket connected to %s", self.name, self.addr)

    def close(self):
        """Close the socket connection."""
        self.socket.close()
        logger.info("%s socket closed", self.name)


def _recv_all(conn, length):
    buffer = bytearray(length)
    view = memoryview(buffer)
    size = 0
    while size < length:
        packet = conn.recv_into(view)
        view = view[packet:]
        size += packet
    return buffer
